import time
import numpy

from numpy import random


def pseudo_random_seed(hyperseed=0):
    '''
    Generate a pseudo random seed based on current time and system random number
    '''
    timestamp = time.time_ns()
    system_random = int(random.random() * 100000000)
    pseudo_random = timestamp + system_random + hyperseed
    
    return pseudo_random % (4294967296)

def weights_and_biases(n_in, n_out):
    avg_in = random.exponential(1.0, size=[n_in])
    avg_out = random.exponential(1.0, size=[n_out])
    weights = numpy.outer(avg_out, avg_in) + random.normal(size=[n_out, n_in])
    weights *= numpy.sqrt(6 / (n_in + n_out))
    bias = random.normal(size=[n_out]) * avg_out
    return weights, bias

class RandomMLP(object):
    '''
    A class for generating random MLPs with given parameters
    '''
    def __init__(self, n_inputs, n_outputs, 
                 n_hidden_layers=None, 
                 activation=None, 
                 seed=None):
        # Set the seed for the random number generator
        if seed is None:
            seed = pseudo_random_seed()
        random.seed(seed)

        # Set the number of hidden units and activation function
        self.hidden_units = [n_inputs]
        if n_hidden_layers is not None:
            if(isinstance(n_hidden_layers, list)):
                self.hidden_units += n_hidden_layers
            elif(isinstance(n_hidden_layers, numpy.ndarray)):
                self.hidden_units += n_hidden_layers.tolist()
            elif(isinstance(n_hidden_layers, tuple)):
                self.hidden_units += list(n_hidden_layers)
            elif(isinstance(n_hidden_layers, int)):
                self.hidden_units.append(n_hidden_layers)
            else:
                raise TypeError(f"Invalid input type of n_hidden_layers: {type(n_hidden_layers)}")
        self.hidden_units.append(n_outputs)
        
        if activation is None:
            self.activation = lambda x:numpy.maximum(0.01*x, x)
        else:
            self.activation = activation
        
        # Initialize weights and biases to random values
        self.weights = []
        self.biases = []
        for i in range(len(self.hidden_units)-1):
            w, b = weights_and_biases(self.hidden_units[i], self.hidden_units[i+1])
            self.weights.append(w)
            self.biases.append(b)
            
    def forward(self, inputs):
        outputs = inputs
        for i, (weight, bias) in enumerate(zip(self.weights, self.biases)):
            if(i >= len(self.weights)-1):
                outputs = self.activation(weight @ outputs + bias)
            else:
                outputs = weight @ outputs + bias
        return outputs
    
    def __call__(self, *args, **kwds):
        return self.forward(*args, **kwds)


"""
Any MDP Task Sampler
"""
import numpy
import gym
import pygame
import time
from numpy import random
from copy import deepcopy
from l3c.utils import pseudo_random_seed, RandomMLP
from l3c.anymdp.solver import check_task_trans, check_task_rewards


def sample_action_mapping(task):
    ndim = task['ndim']
    action_dim = task['action_dim']
    action_map = RandomMLP(action_dim, ndim)

    return {"action_map": action_map}

def sample_observation_mapping(task):
    ndim = task['ndim']
    observation_dim = task['state_dim']
    observation_map = RandomMLP(ndim, observation_dim, n_hidden_layers=random.randint(1, ndim-1))
    return {
        "observation_map": observation_map
    }

def sample_born_loc(task):
    born_loc_num = random.randint(1, 10)
    born_loc = [(random.uniform(-1, 1, size=(task['ndim'],)), 
                random.exponential(1, size=(task['ndim'],))) for i in range(born_loc_num)]
    return {"born_loc": born_loc}

def sample_sgoal(task):
    sgoal_num = random.randint(0, 10)
    sgoal_loc = []
    existing_loc = [loc for loc, _ in task['born_loc']]
    for i in range(sgoal_num):
        min_dist = 0.0
        while min_dist < 0.5:
            sloc = random.uniform(-1, 1, size=(task['ndim'],))
            # calculate the distance between the goal and the born location
            min_dist = 10000   
            for loc in existing_loc:
                dist = numpy.linalg.norm(sloc-loc[0])
                if(dist < min_dist):
                    min_dist = dist
        sink_range = random.uniform(0.02, 0.2)
        reward = random.exponential(5.0)
        sgoal_loc.append((numpy.copy(sloc), sink_range, reward))
        existing_loc.append(numpy.copy(sloc))
    return {"sgoal_loc": sgoal_loc}

def sample_pitfalls(task):
    pf_num = random.randint(20, 100)
    pitfalls_loc = []
    existing_loc = [loc[0] for loc in task['born_loc']]
    if(task['mode']=='sgoal'):
        existing_loc.extend([loc[0] for loc in task['sgoal_loc']])
    for i in range(pf_num):
        min_dist = 0.0
        while min_dist < 0.5:
            pfloc = random.uniform(-1, 1, size=(task['ndim'],))
            # calculate the distance between the goal and the born location
            min_dist = 10000   
            for loc in existing_loc:
                dist = numpy.linalg.norm(pfloc-loc[0])
                if(dist < min_dist):
                    min_dist = dist
        sink_range = random.uniform(0.05, 0.20)
        reward = - random.exponential(1.0)
        pitfalls_loc.append((numpy.copy(pfloc), sink_range, reward))
        existing_loc.append(numpy.copy(pfloc))
    return {"pitfalls_loc": pitfalls_loc}

def sample_line_potential_energy(task):
    num = random.randint(0, 4)
    line_potential_energy = []
    for i in range(num):
        dir = random.uniform(-1, 1, size=(task['ndim'],))
        dir /= numpy.linalg.norm(dir)
        line_potential_energy.append((dir, random.exponential(0.50)))
    return {"line_potential_energy": line_potential_energy}

def sample_point_potential_energy(task):
    num = random.randint(0, 4)
    point_potential_energy = []
    for i in range(num):
        pts = random.uniform(-1, 1, size=(task['ndim'],))
        point_potential_energy.append((pts, random.exponential(0.2), -random.exponential(0.50)))
    return {"point_potential_energy": point_potential_energy}

def sample_dgoal(task, max_order=16, max_item=3):
    num = random.randint(0, 4)
    item_num = random.randint(0, max_item + 1)
    dgoal_loc = [(0, random.normal(size=(task['ndim'], 2)))]
    for j in range(item_num):
        # Sample a cos nx + b cos ny
        order = random.randint(1, max_order + 1)
        factor = random.normal(size=(task['ndim'], 2))
        dgoal_loc.append((order, factor))
    r_range = random.uniform(0.05, 0.40)
    dr = random.exponential(1.0)
    return {"dgoal_loc": dgoal_loc, "dgoal_potential": (r_range, dr)}

def AnyMDPv2TaskSampler(state_dim:int=256,
                 action_dim:int=256,
                 seed=None,
                 verbose=False):
    # Sampling Transition Matrix and Reward Matrix based on Irwin-Hall Distribution and Gaussian Distribution
    # Task:
    # mode: static goal or moving goal
    # ndim: number of inner dimensions
    # born_loc: born location and noise
    # sgoal_loc: static goal location, range of sink, and reward
    # pf_loc: pitfall location, range of sink, and reward
    # line_potential_energy: line potential energy specified by direction and detal_V
    # point_potential_energy: point potential energy specified by location and detal_V


    if(seed is not None):
        random.seed(seed)
    else:
        random.seed(pseudo_random_seed())
    
    task = dict()
    mode = random.choice(["sgoal", "dgoal"])

    task["mode"] = mode
    task["state_dim"] = state_dim
    task["action_dim"] = action_dim
    task["ndim"] = random.randint(3, 33) # At most 32-dimensional space
    task["max_steps"] = random.randint(60, 500) # At most 10-dimensional space
    task["action_weight"] = random.uniform(5.0e-3, 0.10, size=(task['ndim'],))
    task["average_cost"] = - random.exponential(0.01) * random.choice([0, 1])
    task["transition_noise"] = max(0, random.normal(scale=5.0e-3))
    task["reward_noise"] = max(0, random.normal(scale=5.0e-3))

    task.update(sample_observation_mapping(task)) # Observation Model
    task.update(sample_action_mapping(task)) # Action Mapping
    task.update(sample_born_loc(task)) # Born Location

    if(task['mode'] == 'sgoal') :
        task.update(sample_sgoal(task)) # Static Goal Location
    else:
        task.update(sample_dgoal(task)) # Moving Goal Location
    task.update(sample_pitfalls(task)) # Pitfall Location
    task.update(sample_line_potential_energy(task)) # Potential Energy
    task.update(sample_point_potential_energy(task)) # Potential Energy

    return task


"""
Gym Environment For Any MDP
"""
import numpy
import gym
import pygame
import random as rnd
from numpy import random

from gym import error, spaces, utils
from gym.utils import seeding
from l3c.utils import pseudo_random_seed

class AnyMDPEnv(gym.Env):
    def __init__(self, max_steps):
        """
        Pay Attention max_steps might be reseted by task settings
        """
        self.observation_space = spaces.Box(low=-numpy.inf, high=numpy.inf, shape=(1,), dtype=float)
        self.action_space = spaces.Box(low=-1, high=1, shape=(1,), dtype=float)
        self.max_steps = max_steps
        self.task_set = False

    def set_task(self, task):
        for key in task:
            setattr(self, key, task[key])
        # 定义无界的 observation_space
        self.observation_space = gym.spaces.Box(low=-numpy.inf, high=numpy.inf, shape=(self.state_dim,), dtype=numpy.float32)
        # 定义 action_space
        self.action_space = gym.spaces.Box(low=-1, high=1, shape=(self.action_dim,), dtype=numpy.float32)

        self.task_set = True
        self.need_reset = True

    def reset(self):
        if(not self.task_set):
            raise Exception("Must call \"set_task\" first")
        
        self.steps = 0
        self.need_reset = False
        random.seed(pseudo_random_seed())

        loc, noise = rnd.choice(self.born_loc)
        self._inner_state = loc + noise * random.normal(size=self.ndim)
        self._state = self.observation_map(self._inner_state)
        return self._state, {"steps": self.steps}

    def step(self, action):
        if(self.need_reset or not self.task_set):
            raise Exception("Must \"set_task\" and \"reset\" before doing any actions")
        assert numpy.shape(action) == (self.action_dim,)

        ### update inner state
        inner_action = self.action_map(numpy.array(action))
        next_inner_state = self._inner_state + inner_action * self.action_weight + self.transition_noise * random.normal(size=(self.ndim,))

        ### basic reward
        reward = self.average_cost + self.reward_noise * random.normal()
        done = False

        if(self.mode == 'sgoal'):
            # Static Goal (with reset)
            for gs, d, gr in self.sgoal_loc:
                if(numpy.linalg.norm(next_inner_state - gs) < d):
                    done = True
                    reward += gr
                    break
            if(len(self.sgoal_loc) > 0):
                idx = self.steps % len(self.sgoal_loc)
                goal_loc = self.sgoal_loc[idx][0]
            else:
                goal_loc = numpy.zeros_like(self._inner_state)
        else:
            # Dynamic Goal
            dgoal_loc = self.calculate_loc(self.dgoal_loc, self.steps)
            dgoal_dist = numpy.linalg.norm(next_inner_state - dgoal_loc)
            if(dgoal_dist < self.dgoal_potential[0]):
                reward += self.dgoal_potential[1] * (1 - dgoal_dist / self.dgoal_potential[0])
            goal_loc = dgoal_loc

        for gs, d, gr in self.pitfalls_loc:
            if(numpy.linalg.norm(next_inner_state - gs) < d):
                done = True
                reward += gr
                break
        for dir, gr in self.line_potential_energy:
            deta_s = next_inner_state - self._inner_state
            reward += gr * numpy.dot(deta_s, dir)
        for pt, d, gr in self.point_potential_energy:
            pre_s = numpy.linalg.norm(self._inner_state - pt)
            next_s = numpy.linalg.norm(next_inner_state - pt)
            reward += gr * (next_s - pre_s)


        self.steps += 1
        info = {"steps": self.steps, "goal_loc": goal_loc}

        self._inner_state = next_inner_state
        self._state = self.observation_map(self._inner_state)

        done = (self.steps >= self.max_steps or done)
        if(done):
            self.need_reset = True
        return self._state, reward, done, info
    
    def calculate_loc(self, loc, steps):
        # Sample a cos nx + b cos ny
        g_loc = numpy.zeros(self.ndim)
        for n, k in loc:
            g_loc += k[:, 0] * numpy.cos(0.01 * n * self.steps) + k[:, 1] * numpy.sin(0.01 * n * self.steps)
        return g_loc / len(loc)
    
    @property
    def state(self):
        return numpy.copy(self._state)
    
    @property
    def inner_state(self):
        # 复制内部状态
        return numpy.copy(self._inner_state)


if __name__=="__main__":
    import gym
    from l3c.anymdpv2 import AnyMDPv2TaskSampler

    task = AnyMDPv2TaskSampler(state_dim=128, 
                             action_dim=16)
    max_steps = 5000
    prt_freq = 100

    # Test Random Policy
    env = gym.make("anymdp-v2-visualizer")
    env.set_task(task)
    state, info = env.reset()
    acc_reward = 0
    epoch_reward = 0
    done = False
    obs_arr = []
    act_arr = []
    state_arr = []

    steps = 0
    while steps < max_steps:
        action = env.action_space.sample()
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
    print("Random Policy Summary: {}".format(acc_reward))
    env.visualize_and_save()

    print("Test Passed")


import numpy
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from sklearn.datasets import load_digits
from .anymdp_env import AnyMDPEnv
from l3c.utils import pseudo_random_seed

class AnyMDPv2Visualizer(AnyMDPEnv):
    
    def set_task(self, task):
        # Set task will automatically reset all records
        super().set_task(task)
        self.observation_records = []
        self.inner_state_records = []
        self.action_records = []
        self.reward_records = []
        self.goal_records = []
        self.colors = []

        self.color_spec_type = [['green', 'black', 'red'],
                              ['green', 'blue', 'red']]
    
    def color_spec(self, i):
        return [self.color_spec_type[i][idx] for idx in self.colors]

    def reset(self):
        obs, info = super().reset()
        self.observation_records.append(numpy.copy(obs))
        self.action_records.append(numpy.zeros((self.action_dim,)))
        self.inner_state_records.append(numpy.copy(self.inner_state))
        self.reward_records.append(0.0)
        self.goal_records.append(numpy.zeros((self.ndim, )))
        self.colors.append(0)

        return obs, info
    
    def step(self, action):
        obs, reward, done, info = super().step(action)
        self.observation_records.append(numpy.copy(obs))
        self.inner_state_records.append(numpy.copy(self.inner_state))
        self.goal_records.append(numpy.copy(info["goal_loc"]))
        self.action_records.append(numpy.copy(action))
        self.reward_records.append(reward)
        if(done):
            self.colors.append(2)
        else:
            self.colors.append(1)
        return obs, reward, done, info

    def visualize_and_save(self, filename=None):
        tsne = TSNE(n_components=2, random_state=pseudo_random_seed(),
                    perplexity=10, max_iter=500, learning_rate=100)
        if(filename is not None):
            file_name = file_name
        else:
            file_name = "./anymdp_visualizer_output.pdf"
        obs_arr = numpy.array(self.observation_records, dtype="float32")
        act_arr = numpy.array(self.action_records, dtype="float32")
        pitfalls = [x[0] for x in self.pitfalls_loc]
        s_arr = numpy.array(self.inner_state_records + self.goal_records + pitfalls, dtype="float32")
        max_steps = len(self.inner_state_records)

        obs_tsne = tsne.fit_transform(numpy.array(obs_arr))
        act_tsne = tsne.fit_transform(numpy.array(act_arr))
        s_tsne = tsne.fit_transform(numpy.array(s_arr))
        
        c1 = self.color_spec(0)
        c2 = self.color_spec(1)

        plt.figure(figsize=(10, 8))
        plt.subplot(2, 2, 1)
        scatter = plt.scatter(obs_tsne[:, 0], obs_tsne[:, 1], c=c1, s=10, alpha=0.2)
        plt.title("Observation", fontsize=12, fontweight='bold', color='blue', pad=10)
        plt.subplot(2, 2, 2)
        scatter = plt.scatter(act_tsne[:, 0], act_tsne[:, 1], c=c1, s=10, alpha=0.2)
        plt.title("Action", fontsize=12, fontweight='bold', color='blue', pad=10)
        plt.subplot(2, 2, 3)
        scatter = plt.scatter(s_tsne[:max_steps, 0], s_tsne[:max_steps, 1], c=c1, label='Agent', s=10, alpha=0.2, marker='o')
        label = f'Static Goal {len(self.sgoal_loc)}' if(self.mode == 'sgoal') else 'Dynamic Goal'
        scatter = plt.scatter(s_tsne[max_steps:2*max_steps, 0], s_tsne[max_steps:2*max_steps, 1], c=c2, label=label, s=10, alpha=1.0, marker='+')
        scatter = plt.scatter(s_tsne[2*max_steps:, 0], s_tsne[2*max_steps:, 1], c='brown', label='pitfalls', s=10, alpha=1.0, marker='D')
        plt.legend()
        plt.title("Inner States", fontsize=12, fontweight='bold', color='blue', pad=10)
        plt.subplot(2, 2, 4)
        scatter = plt.plot(numpy.arange(numpy.size(self.reward_records)), self.reward_records, c='red')
        plt.title("Reward", fontsize=12, fontweight='bold', color='blue', pad=10)
        plt.savefig(file_name)


"""
AnyMDP Task Visualization
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.colors as mcolors
import matplotlib.cm as cm
 

def task_visualizer(task, show_gui=True, file_path=None):
    if("state_embedding" not in task):
        raise Exception("No state embedding found in task")
    
    ns, sv = task["state_embedding"].shape
    ns1, na, ns2 = task["reward"].shape
    ns3, na1, ns4 = task["transition"].shape

    assert ns == ns1 == ns2 == ns3 == ns4
    assert na == na1

    projection_matrix = np.random.rand(sv, 3)

    s= task["state_embedding"]
    s_min = s.min(axis=0)
    s_max = s.max(axis=0)
    coordinates = np.matmul((s - s_min) / (s_max - s_min), projection_matrix)
    link_strength_normalized = np.mean(task["transition"], axis=1)
    rewards = np.mean(task["reward"], axis=(0, 1))

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    norm = mcolors.Normalize(vmin=np.min(rewards), vmax=np.max(rewards))
    cmap = cm.viridis

    ax.scatter(coordinates[:, 0], coordinates[:, 1], coordinates[:, 2], 
                         c=cmap(norm(rewards)), marker='o')

    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cbar = fig.colorbar(sm, ax=ax, shrink=0.2, aspect=5)
    cbar.set_label('Rewards')

    for i in range(ns):
        for j in range(ns):
            if i != j and link_strength_normalized[i,j] > 0.1:  
                x_values = [coordinates[i, 0], coordinates[j, 0]]
                y_values = [coordinates[i, 1], coordinates[j, 1]]
                z_values = [coordinates[i, 2], coordinates[j, 2]]
                c = max(0, float(1 - 2 * link_strength_normalized[i, j]))
                ax.plot(x_values, y_values, z_values, color=(c,c,c), linewidth=2)
       
    ax.set_title('AnyMDP Task Visualization')

    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_zticklabels([])
    
    ax.xaxis.set_ticks_position('none')
    ax.yaxis.set_ticks_position('none')
    ax.zaxis.set_ticks_position('none')

    ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    
    ax.grid(False)
    if show_gui:
        plt.show()
    if file_path is not None:
        plt.savefig(file_path)


if __name__ == '__main__':
    from l3c.anymdp import AnyMDPTaskSampler
    task = AnyMDPTaskSampler(128, 5, keep_metainfo=True)
    task_visualizer(task)